/*
 * pp65.c : A 24C/LC65 SEEPROM programmer
 *
 * Usage:
 *
 *      pp65  [ -rpv23! ]  binfile
 *
 * Where binfile is the data in binary format.
 *
 * Switches (options) are introduced by - (or /) and are encoded as
 * a sequence of one or more characters from the set {rpv23!}.
 * Switches can be given individually (like -r /2 -!) or in groups
 * (like -r2! or /r!2 or /2/r/! which are all synonyms).  Switches are
 * recognized anywhere in the command line (but don't use the DOS-like  
 * pp65/r/!/2  form - this should be written  pp65 /r/!/2;  space around
 * switch groups is important).
 *
 * Switches:
 *    o    p           -  programs SEEPROM from binary file (default)
 *    o    r           -  reads SEEPROM into binary file
 *    o    v           -  verifies SEEPROM againts binary file
 *
 * Ports:
 *    o   2, 3         -  hardware on LPT2 or LPT3 (default is LPT1).
 *
 * Misc:
 *    o   !            -  start programming immediately (default is
 *                        to wait for a key press)
 *
 * E.g., -v2 means verify the SEEPROM against binary file and assume the
 * hardware is on LPT2.
 *
 *
 * Notes/restrictions/bugs:
 *
 *      o  This program needs external hardware to be attached to a
 *         parallel port (see david.txt or henk.txt for an ASCII schematic).
 *      o  Don't use from Windows; another program could grab the LPT port.
 *      o  Run pp65, pp65 -2 or pp65 -3 to initialize the LPT port.
 *      o  Programming sometimes gives verify errors.  Just program
 *         again and things will do fine!
 *      o  This is NOT a production quality programmer. (No surprise!).
 *
 *
 * Revision history:
 *
 * 1995-06-24:  v0.0  - first hack from David Tait's pp.c code
 * 1995-06-26:  v0.1b - first beta release to the public
 *
 * This version (v0.1b) is a beta release.  If you receive this program
 * consider yourself a beta tester (of course, you are under no
 * obligation).  I will remove beta status when 10 (arbitrary
 * I know) testers tell me they have used it to successfully program a
 * SEEPROM using the hardware in david.txt or henk.txt.
 *
 *
 * To do:
 *
 *      o You tell me!
 *
 *
 * Acknowledgements:
 *
 * The programming algorithms used are taken from the data sheet
 * DS21058B (C) 1993 Microchip Technology Incorporated.  Thanks
 * to Arizona Microchip Technology Ltd for making this information
 * available without restriction.
 *
 *
 * Legal stuff:
 *
 * This is copyright software, but I do not seek to profit from
 * it.  I wrote it in the hope that it will prove useful.  Feel
 * free to modify the program, but please DO NOT distribute
 * modified versions.  Instead, send me your modifications and
 * improvements so that they may, at my discretion, be included in a
 * future release.
 *
 * Although I cannot undertake to support this software, I would
 * welcome bug reports, suggestions or comments which should be
 * sent to:
 *
 *    Luca Bertagnolio
 *    E-mail: <berta@dsi.unimi.it>
 *
 * Copyright (C) 1995 David Tait.
 * This program is free software.  Permission is granted to use,
 * copy, or redistribute this program so long as it is not sold
 * for profit.
 *
 * THIS PROGRAM IS PROVIDED AS IS AND WITHOUT WARRANTY OF ANY KIND,
 * EITHER EXPRESSED OR IMPLIED.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <conio.h>

#define IN      128

#define VDD     8
#define VPP     4
#define CLK     2
#define OUT     1

#define TDLY  0
#define TSET  0      /* the units for these are CPU dependent */
#define THLD  0      /* I doubt if they are necessary even    */
#define TDLY  0      /* on the fastest PC, but ...            */

#define PROGRAM 0
#define DATA    1
#define DSIZE   8192

#define READ    0xa1
#define WRITE   0xa0

#define inbit           (~inportb(s_reg)&IN)
#define vppoff          (d_bits &= ~VPP)
#define vppon           (d_bits |= VPP)
#define vddon           (d_bits &= ~VDD)
#define vddoff          (d_bits |= VDD)
#define clklo           (d_bits &= ~CLK)
#define clkhi           (d_bits |= CLK)
#define outlo           (d_bits &= ~OUT)
#define outhi           (d_bits |= OUT)
#define assert          (outportb(d_reg,d_bits))


int databuf[DSIZE];
int PRGDLY=4;		/* delay in programming loop */

int read_data=0,program_data=0,verify_data=0;

int d_bits = 0;
int d_reg;
int s_reg;
int check;
int iport = 0;
int wait = 1;

char *progname = "24C/LC65 Programmer";
char *version = "Version 0.0";

void tiny_delay(int ticks)
{
   while ( ticks-- )
      ;
}

void alarm(void)
{
    sound(300);delay(20);
    sound(400);delay(20);
    sound(500);delay(20);nosound();}

void idle_mode(void)
{
   vppoff, clkhi, outhi, assert;
   tiny_delay(TDLY);
   vddoff, assert;
}

void start_mode(void)
{
   vddon, clkhi, outhi, assert;
   tiny_delay(TDLY);
   outlo, assert;
   tiny_delay(TDLY);
   clklo, assert;
}

void stop_mode(void)
{
   outlo, assert;
   tiny_delay(TDLY);
   clkhi, assert;
   tiny_delay(TDLY);
   outhi, assert;
   tiny_delay(TDLY);
}

void quit(char *s)
{
   printf("\n\rpp65: %s\n\r",s);
   idle_mode();
   exit(1);
}

int c_break(void)
{
   quit("Control-Break pressed.  Program aborting ...\n");
   return(0);
}


void usage(void)
{
   printf("\r%s  %s\n\n",progname,version);
   printf("\rUsage: pp65  [ -prv23! ]  binfile\n\n");
   printf("\r       binfile = Binary data file\n\n");
   printf("\r       Action:   p = program, r = read, v = verify\n");
   printf("\r       Port:     2 = LPT2, 3 = LPT3\n");
   printf("\r       Misc:     ! = no wait\n\n");
   printf("\r       Defaults: program, LPT1, wait\n\n");
   idle_mode();
   exit(1);
}

void test_hw(void)
{
    int b;

    vddon, outhi, assert;   /* better have VDD on in case PIC is there */
    tiny_delay(TDLY);
    b = inbit;
    outlo, assert;
    if ( b != IN || inbit != 0 ) {
       printf("\rpp65: Using LPT at 0x%04X\n",d_reg);
       quit("Hardware fault.  Check power, connections and port used.");
    }
    idle_mode();
}

void setup(void)
{
   int i, b;

   vppoff, vddoff, clklo, outlo, assert;
   d_reg = peek(0,0x408+iport);  /* find the base address of LPT port */
   s_reg = d_reg+1;

   switch ( d_reg ) {  /* I don't want to blow anything up, so check port */
       case 0x3BC:
       case 0x378:
       case 0x278: break;
       default:  printf("\rpp65: LPT at 0x%04X\n",d_reg);
		 quit("Unlikely LPT address");
   }


   test_hw();

   for ( i=0; i<DSIZE; ++i )
      databuf[i] = 0xFF;
}

void clock_out(int bit)
{
   bit? outhi: outlo, clklo, assert;
   tiny_delay(TSET);
   clkhi, assert;
   tiny_delay(THLD);
   clklo, assert;
}

int clock_in(void)
{
   int b;

   outhi, clklo, assert;
   tiny_delay(TSET);
   clkhi, assert;
   b = inbit? 1: 0;
   tiny_delay(THLD);
   return b;
}

int in_ack(void)
{
   int b;

   b = clock_in();
   clklo,assert;
   return b;
}

void out_ack(void)
{
   clock_out(0);
}

void out_byte(int w)
{
   int b;

   for ( b=0; b<8; b++ ) {
     clock_out((w>>7)&0x1);
     w=w<<1;
   }
}

int in_byte(void)
{
   int b, w;

   for ( w=0, b=0; b<8; ++b )
     w += clock_in()<<(7-b);
   return w;
}

void get_option(char *s)
{
   char *sp;

   for ( sp=s; *sp; ++sp )
      switch ( *sp ) {
	 case 'p':
	 case 'P': program_data = 1; break;
	 case 'r':
	 case 'R': read_data = 1; break;
	 case 'v':
	 case 'V': verify_data = 1; break;
	 case '2': iport = 2; break;
	 case '3': iport = 4; break;
	 case '!': wait = 0; break;
	 case '-':
	 case '/': break;
	 default: usage();
      }
}

void loadbin(FILE *fp,int buf[], int bufsize)
{
   unsigned address;

   for (address=0; address<bufsize; ++address) {
     buf[address] = getc(fp);
   }

   fclose(fp);
}

void savebin(FILE *fp,int buf[], int bufsize)
{
   unsigned address;

   for (address=0; address<bufsize; ++address) {
     putc(buf[address],fp);
   }

   fclose(fp);
}

void program_all(void)
{
   unsigned address;
   int i,t=0;
   char wait_ch[4]={'-','\\','|','/'};

   if ( wait ) {
      printf("\nInsert SEEPROM and press a key to start (^C to abort) ...");
      if ( ((i=getch()) == 3) || i==27 )
	 quit("Aborted");
   }
   printf("\nProgramming ... ");

   for (address=0; address<DSIZE; ++address) {
     start_mode();
     out_byte(WRITE);
     in_ack();
     out_byte(address/256);
     in_ack();
     out_byte(address%256);
     in_ack();
     out_byte(databuf[address]);
     in_ack();
     stop_mode();
     putch(wait_ch[t++]);
     gotoxy(wherex()-1,wherey());
     if (t==4) t=0;
     delay(PRGDLY);
   }

   putch(' ');

}

void verify_all(void)
{
   unsigned address;
   int i,num;

   if ( wait ) {
      printf("\nInsert SEEPROM and press a key to start (^C to abort) ...");
      if ( ((i=getch()) == 3) || i==27 )
	 quit("Aborted");
   }
   printf("\nVerifying ...");

   start_mode();
   out_byte(WRITE);
   in_ack();
   out_byte(0x00);
   in_ack();
   out_byte(0x00);
   in_ack();
   start_mode();
   out_byte(READ);
   in_ack();

   for (num=0,address=0; address<DSIZE; ++address) {
     i = in_byte() & 0xff;
     out_ack();
     if (databuf[address] != i) {
       num++;
       printf("\n%04x: %02x != %02x",address,i,databuf[address]);
     }
   }

   if (num>0) {
     printf("\nVerify failed with %i errors!\n",num);
     alarm();
     }
   else printf("\nVerify successful!\n");

   stop_mode();
}

void read_all(void)
{
   unsigned address;
   int i;

   if ( wait ) {
      printf("\nInsert SEEPROM and press a key to start (^C to abort) ...");
      if ( ((i=getch()) == 3) || i==27 )
	 quit("Aborted");
   }
   printf("\nReading ...");

   start_mode();
   out_byte(WRITE);
   in_ack();
   out_byte(0x00);
   in_ack();
   out_byte(0x00);
   in_ack();
   start_mode();
   out_byte(READ);
   in_ack();

   for (address=0; address<DSIZE; ++address) {
     databuf[address] = in_byte() & 0xff;
     out_ack();
   }

   stop_mode();
}

void main(argc, argv)
int argc;
char *argv[];
{
    FILE *binfp;
    int i, c;

    for (i=1; i<argc; ++i) {
       if ( (c = *argv[i]) == '-' || c == '/' )
	 get_option(++argv[i]);
       else {
	 binfp = fopen(argv[i],"rb");
	 break;
       }
    }

    if ((argv[i]==NULL)||(++i<argc)||(argc==1)) usage();

    setup();

    if (!(read_data||verify_data)) program_data=1;
    if (read_data&&verify_data)    quit("conflicting parameters!");
    if (read_data&&program_data)   quit("conflicting parameters!");
    if (program_data&&verify_data) quit("conflicting parameters!");

    if (read_data&&(binfp!=NULL))  quit("output file exists!");

    if (program_data||verify_data) {
      if (binfp==NULL) quit("input file not found!");
      loadbin(binfp,databuf,DSIZE);
      fclose(binfp);
    }

    ctrlbrk(c_break);

    if (read_data) {
      printf("READ mode\n");
      fclose(binfp);
      binfp = fopen(argv[--i],"wb");

      read_all();
      idle_mode();

      savebin(binfp,databuf,DSIZE);

    }

    if (program_data) {
      printf("PROGRAM mode\n");

      program_all();
      idle_mode();

      wait=0;
      verify_all();
      idle_mode();

    }

    if (verify_data) {
      printf("VERIFY mode\n");

      verify_all();
      idle_mode();

    }

}
